
// HopalongSaver.cpp
// Implementation for the HopalongSaver class

// HopalongSaver, a screensaver module for BeOS
// Jon Pimble -- fatrat@fat-rat.com



#include "HopalongSaver.h"



extern "C" _EXPORT BScreenSaver *instantiate_screen_saver( BMessage *msg, image_id id )
{
  return new HopalongSaver( msg, id ); 
} 



// Constructor
HopalongSaver::HopalongSaver( BMessage *archive, image_id image ) 
          : BScreenSaver( archive, image ) 
{ 
  // Get initial values from screensaver caller, and make sure they're
  //  within an acceptable range; if not, use default values
  if(archive->FindInt32("ticks", &ticks) != B_OK)
  {
    ticks = DEFAULT_TICKS;
  }
  if((ticks < MIN_TICKS) || (ticks > MAX_TICKS))
  {
    ticks = DEFAULT_TICKS;
  }
  old_ticks = ticks;
  
}



// Saves the state of the screensaver (stores config varliables)
status_t HopalongSaver::SaveState(BMessage *into) const
{
  // Save state of configuration settings
  into->AddInt32("ticks", ticks);
  
  return B_OK;
}



void HopalongSaver::StartConfig(BView *view) 
{ 
  // Create and attach credits/oonfig view
  // Note: We don't have to keep track of the object we create, because it will be
  //  destroyed for us when its parent view is destroyed
  ConfigView *credits;
  
  credits = new ConfigView(view->Bounds(), "Credits", &ticks);

  view->AddChild(credits);
}



// Setup screensaver parameters before inital drawing
status_t HopalongSaver::StartSaver( BView *view, bool preview ) 
{
  // Time between calls to Draw(), in microseconds (1000000us == 1s)
  SetTickSize(ticks);

  // Make a note of whether we're in the preview mode
  is_preview = preview;
  
  // Precalculations used for positioning and scaling the fractal
  scale_width = view->Bounds().Width()/500.0;
  scale_height = view->Bounds().Height()/500.0;
  
  // Seed random number generator 
  srand(time(NULL));
  
  // Must return B_OK for screensaver to start
  return B_OK;
}
       


// Main drawing function
void HopalongSaver::Draw(BView *view, int32 frame)
{
  static BRect tmp_rect;
  
  // If any config variables changed, apply them
  //  (This is necessary to keep the saver in sync with changes made in the ConfigView)
  if(ticks != old_ticks)
  {
    old_ticks = ticks;
    SetTickSize(ticks);
  }
  
  // In the first frame, blank the background, set parameters for fractal
  if( (frame == 0)
   || ( (!is_preview) && (iter_count > MAX_ITERS) )
   || ( (is_preview) && (iter_count > PREVIEW_MAX_ITERS) )
  )
  {  
    // Black out background
    view->SetDrawingMode(B_OP_COPY);
    view->SetHighColor(0, 0, 0);
    view->FillRect(view->Bounds());

    // Setup new fractal at origin, with random paramaters
    hopper.x = hopper.y = 0.0;

    hopper.a = (float) (rand() % 10000 - rand() % 10000)/100.0;
    hopper.b = (float) (rand() % 10000 - rand() % 10000)/100.0;
    hopper.c = (float) (rand() % 10000 - rand() % 10000)/100.0;
    
    // Choose a random color to start at
    hsv_draw_color.h = rand() % 360;
    hsv_draw_color.s = 0.5 + (float) (rand() % 50)/100.0;
    hsv_draw_color.v = 0.5 + (float) (rand() % 50)/100.0;
    
    last_hue = hsv_draw_color.h;
    last_rgb_color = hsv_to_rgb(hsv_draw_color);

    iter_count = 0;
    
    // Random offset from the center of the screen
    center_x = ((rand() % (int) view->Bounds().Width())/2) - ((rand() % (int) view->Bounds().Width())/2) + (view->Bounds().Width()/2.0);
    center_y = ((rand() % (int) view->Bounds().Height())/2) - ((rand() % (int) view->Bounds().Height())/2) + (view->Bounds().Height()/2.0);

    // We blend the newly plotted points for a smoother look
    view->SetDrawingMode(B_OP_BLEND);
  }
  
  // Preview mode has a different step rate, to make it animate quickly
  if(is_preview)
  {
    hsv_draw_color.h += PREVIEW_STEPRATE;
  }
  else
  {
    hsv_draw_color.h += STEPRATE;
  }
    

  // Simplified version of formula  
  float tmp_y = hopper.a - hopper.x;
  if(hopper.x < 0)
  {
    hopper.x = hopper.y + sqrt(fabs((hopper.b*hopper.x)-hopper.c));
  }
  else
  {
    hopper.x = hopper.y - sqrt(fabs((hopper.b*hopper.x)-hopper.c));
  }  
  hopper.y = tmp_y;
  

  // Calculate color only if we've moved at least one tenth of a dregree around the
  //  color wheel
  if((hsv_draw_color.h - 0.1) > last_hue)
  {
    last_rgb_color = hsv_to_rgb(hsv_draw_color);
    last_hue = hsv_draw_color.h;
    view->SetHighColor(last_rgb_color);
  }
  
  // Draw the point for this iteration
  tmp_rect.left = hopper.x*scale_width + center_x;
  tmp_rect.top =  hopper.y*scale_height + center_y;
  tmp_rect.right = tmp_rect.left;
  tmp_rect.bottom = tmp_rect.top;

  view->FillRect(tmp_rect);

  // Keep track of the number of iterations performed so far
  iter_count++;  
}



// Convert an HSV color value to an RGB one
rgb_color HopalongSaver::hsv_to_rgb(hsv_color & hsv_to_convert)
{
  float f, p1, p2, p3, scaled_hue;
  float red, green, blue;
  rgb_color rgb_dest_color;
  hsv_color hsv_scratch = hsv_to_convert;
    
  // Be sure that hue is in the range we want (0 to 359.9999)
  hsv_scratch.h = fabs(fmodf(fabs(hsv_scratch.h), 360));
    
  // Be sure that s, v are within acceptable range (0 to 1.0)
  hsv_scratch.s = fabs(fmodf(hsv_scratch.s, 1.0));
  hsv_scratch.v = fabs(fmodf(hsv_scratch.v, 1.0));
  
  
  // We want to deal with a hue value between 0 and 6 (corners on the HSV
  //  hexagon model)
  scaled_hue = hsv_scratch.h / 60.0;
    
  // Scale hue to a value between 0 and 1.0
  f = scaled_hue - floor(scaled_hue);

  // Coordinate conversion from HSV hexagon to RGB cube
  p1 = hsv_scratch.v * (1.0 - hsv_scratch.s);
  p2 = hsv_scratch.v * (1.0 - (hsv_scratch.s * f));
  p3 = hsv_scratch.v * (1.0 - (hsv_scratch.s * (1.0 - f)));


  // Pick RGB values based on coordinates in HSV hexagon
  switch((int) floor(scaled_hue))
  {
    case 0:
      red = hsv_scratch.v;
      green = p3;
      blue = p1;
      break;
        
    case 1:
      red = p2;
      green = hsv_scratch.v;
      blue = p1;
      break;
        
    case 2:
      red = p1;
      green = hsv_scratch.v;
      blue = p3;
      break;
        
    case 3:
      red = p1;
      green = p2;
      blue = hsv_scratch.v;
      break;
        
     case 4:
      red = p3;
      green = p1;
      blue = hsv_scratch.v;
      break;
        
    case 5:
      red = hsv_scratch.v;
      green = p1;
      blue = p2;
      break;
        
    // If we made it to the default case, something bad happened!
    //  Set the rgb_color values to 0,0,0 as default
    default:
      red = 0;
      green = 0;
      blue = 0;
      break;
  }
  
  
  // Scale color values to 0, 255 for the color_rgb values
  rgb_dest_color.red = (uint8) (red * 255.0);
  rgb_dest_color.green = (uint8) (green * 255.0);
  rgb_dest_color.blue = (uint8) (blue * 255.0);
  rgb_dest_color.alpha = (uint8) (hsv_scratch.alpha * 255.0);

  // Return converted color
  return rgb_dest_color;
}
